// See the end of this file for copyright and license terms.

use std::str::Chars;

use self::cursor::Cursor;

mod cursor;

#[derive(Clone)]
pub struct Lexer<'a> {
    cursor: Cursor<'a>,
    token_line: usize,
    token_col: usize,
    has_ended: bool,
}

#[derive(Debug)]
pub struct Token {
    pub kind: TokenKind,
    pub raw: String,
    pub line: usize,
    pub col: usize,
}

impl Token {
    pub fn require_kind(&self, kind: TokenKind) -> Result<&Token, SyntaxError> {
        if self.kind == kind {
            Ok(self)
        } else {
            Err(SyntaxError {
                msg: format!("Expected {:?}, got {:?}", kind, self),
                line: self.line,
                col: self.col,
            })
        }
    }
}

#[derive(Debug, PartialEq)]
pub enum TokenKind {
    EOF,
    Semi,
    Eq,
    Minus,
    TrueKeyword,
    FalseKeyword,
    Ident,
    IntLiteral,
    StringLiteral,
    Whitespace,
}

impl Iterator for Lexer<'_> {
    type Item = Result<Token, SyntaxError>;

    fn next(&mut self) -> Option<Result<Token, SyntaxError>> {
        if self.has_ended {
            None
        } else {
            match self.stupid_wrapper_for_next() {
                Some(Ok(token)) if token.kind == TokenKind::Whitespace => self.next(),
                None => {
                    self.has_ended = true;
                    Some(Ok(self.make_token(TokenKind::EOF)))
                },
                x => x
            }
        }
    }
}

impl<'a> Lexer<'a> {
    pub const fn new(stream: Chars<'a>) -> Lexer<'a> {
        Lexer {
            cursor: Cursor::new(stream),
            token_line: 1,
            token_col: 1,
            has_ended: false,
        }
    }

    pub fn syntax_error<T>(&self, msg: String) -> Result<T, SyntaxError> {
        Err(SyntaxError {
            msg,
            line: self.line(),
            col: self.col(),
        })
    }

    pub fn line(&self) -> usize {
        self.cursor.line()
    }

    pub fn col(&self) -> usize {
        self.cursor.col()
    }

    pub fn peek(&self) -> Option<Result<Token, SyntaxError>> {
        self.clone().next()
    }

    fn stupid_wrapper_for_next(&mut self) -> Option<Result<Token, SyntaxError>> {
        Some(match self.cursor.next()? {
            c if is_whitespace(c) => Ok(self.read_whitespace()),

            ';' => Ok(self.make_token(TokenKind::Semi)),

            '=' => Ok(self.make_token(TokenKind::Eq)),

            '-' => Ok(self.make_token(TokenKind::Minus)),

            '"' => self.read_string(),

            c @ '1'..='9' => Ok(self.read_int_rest(c)),

            c @ 'a'..='z' | c @ 'A'..='Z' => Ok(self.read_keyword_or_ident(c)),

            c => self.syntax_error(format!("Unexpected token {:?}", c)),
        })
    }

    fn read_keyword_or_ident(&mut self, first: char) -> Token {
        let kind = match first {
            't' if self.skip_if_match("rue") => TokenKind::TrueKeyword,
            'f' if self.skip_if_match("alse") => TokenKind::FalseKeyword,
            _ => {
                self.cursor.next_until(|c| !is_ident_part(c));
                TokenKind::Ident
            },
        };

        self.make_token(kind)
    }

    fn read_string(&mut self) -> Result<Token, SyntaxError> {
        let mut s = String::new();

        loop {
            match self.cursor.next() {
                None | Some('"') => break,

                Some('\\') => match self.read_escape_seq() {
                    Ok(c) => s.push(c),
                    Err(e) => return Err(e),
                },

                Some('\n') => return self.syntax_error(String::from("Unexpected EOL")),

                Some(c) => s.push(c),
            }
        }

        self.cursor.chop();
        Ok(self.make_token_raw(TokenKind::StringLiteral, s))
    }

    fn read_escape_seq(&mut self) -> Result<char, SyntaxError> {
        match self.cursor.next() {
            Some('\\') => Ok('\\'),
            Some('n') => Ok('\n'),
            Some('"') => Ok('"'),
            _ => self.syntax_error(String::from("Illegal escape sequence")),
        }
    }

    fn read_int_rest(&mut self, first: char) -> Token {
        self.cursor.next_until(|c| !matches!(c, '0'..='9'));
        self.make_token(TokenKind::IntLiteral)
    }

    fn read_whitespace(&mut self) -> Token {
        self.cursor.next_until(|c| !is_whitespace(c));
        self.make_token(TokenKind::Whitespace)
    }

    fn skip_if_match(&mut self, s: &str) -> bool {
        let mut chars = self.cursor.chars();

        for c in s.chars() {
            if chars.next() != Some(c) {
                return false;
            }
        }

        self.cursor
            .advance_by(s.len())
            .expect("this shouldn't be possible lol");
        true
    }

    fn make_token_raw(&mut self, kind: TokenKind, raw: String) -> Token {
        let t = Token {
            kind,
            raw,
            line: self.token_line,
            col: self.token_col,
        };

        self.token_line = self.cursor.line();
        self.token_col = self.cursor.col();

        t
    }

    fn make_token(&mut self, kind: TokenKind) -> Token {
        let s = self.cursor.chop();
        self.make_token_raw(kind, s)
    }
}

fn is_whitespace(c: char) -> bool {
    matches!(c, ' ' | '\n' | '\t')
}

fn is_ident_part(c: char) -> bool {
    matches!(c, 'A'..='Z' | 'a'..='z' | '0'..='9' | '_')
}

pub struct SyntaxError {
    pub msg: String,
    pub line: usize,
    pub col: usize,
}

// Copyright (C) 2021 Fefie <owo@fef.moe>
// This file is part of gayconf.
//
// gayconf is non-violent software: you can use, redistribute,
// and/or modify it under the terms of the CNPLv6+ as found
// in the LICENSE file in the source code root directory or
// at <https://git.pixie.town/thufie/CNPL>.
//
// gayconf comes with ABSOLUTELY NO WARRANTY, to the extent
// permitted by applicable law.  See the CNPL for details.
